--------------------------  SETUP ----------------------------
USE AdventureWorks2014;

-- Create demo schema
IF NOT EXISTS ( SELECT  *
                FROM    sys.schemas
                WHERE   name = N'Demo' ) 
    EXEC('CREATE SCHEMA [Demo] AUTHORIZATION [dbo]');
GO

DROP SECURITY POLICY IF EXISTS [dbo].[NameFilter]
DROP FUNCTION IF EXISTS [dbo].[fn_firstletter]
DROP TABLE IF EXISTS AdventureWorks2014.Demo.Person; 

SET STATISTICS TIME ON
SET STATISTICS IO ON

-------------------------- Heap versus Table ----------------------------
--Create a heap
SELECT *
INTO [AdventureWorks2014].[Demo].[Person]
FROM [AdventureWorks2014].[Person].[Person]


--Table Scan
SELECT * FROM [AdventureWorks2014].[Demo].[Person]

--Add a primary key
ALTER TABLE [AdventureWorks2014].[Demo].[Person] 
ADD  CONSTRAINT [PK_Demo_Person_BusinessEntityID] PRIMARY KEY CLUSTERED 
(
	[BusinessEntityID] ASC
)

--Clustered Index Scan
SELECT * FROM [AdventureWorks2014].[Demo].[Person]

--Clustered Index Seek
SELECT * FROM [AdventureWorks2014].[Demo].[Person] 
WHERE BusinessEntityID = 25


--------------------------  Performance comparison ----------------------------

DBCC TRACEON (652,-1);
--Comparing performance
--Drop Index
ALTER TABLE [AdventureWorks2014].[Demo].[Person] 
DROP CONSTRAINT [PK_Demo_Person_BusinessEntityID]

--Empty out buffer pool
CHECKPOINT
DBCC DROPCLEANBUFFERS

SELECT * FROM [AdventureWorks2014].[Demo].[Person] 
WHERE BusinessEntityID = 25



--Add index back in
ALTER TABLE [AdventureWorks2014].[Demo].[Person] 
ADD  CONSTRAINT [PK_Demo_Person_BusinessEntityID] PRIMARY KEY CLUSTERED 
(
	[BusinessEntityID] ASC
)
CHECKPOINT
DBCC DROPCLEANBUFFERS

SELECT * FROM [AdventureWorks2014].[Demo].[Person] 
WHERE BusinessEntityID = 25

DBCC TRACEOFF (652,-1);

--------------------------  Extra Lookups ----------------------------

--Search by non-primary key
--Clustered Index Scan
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName = N'Sanchez'


--Add a nonclustered index
CREATE NONCLUSTERED INDEX [IX_Demo_Person_LastName_FirstName_MiddleName] 
ON [AdventureWorks2014].[Demo].[Person]
(
	[LastName] ASC,
	[FirstName] ASC,
	[MiddleName] ASC
)

--Search by non-primary key
--Nonclustered Index Seek
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName = N'Sanchez'

--Nonclustered Index Scan
SELECT COUNT(*), Lastname
FROM [AdventureWorks2014].[Demo].[Person]
GROUP BY LastName

--SARGability
--Index Scan
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LTRIM(LastName) = N'Sanchez'

--SARGability 2
--Index Seek
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName = LTRIM(N'Sanchez')


--SARGability 3
--Index Scan
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName LIKE N'%Sanchez'

--SARGability 4
--Index Seek
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName LIKE N'Sanchez%'

--Key Lookup
SELECT *
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName = N'Sanchez'

--Drop Index
ALTER TABLE [AdventureWorks2014].[Demo].[Person] DROP CONSTRAINT [PK_Demo_Person_BusinessEntityID]

--RID Lookup
SELECT *
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName = N'Sanchez'

--Nested Loop and Hash Match
SELECT e.BusinessEntityID, e.HireDate, d.Name
  FROM [AdventureWorks2014].[HumanResources].[Employee] e
  INNER JOIN [AdventureWorks2014].[HumanResources].[EmployeeDepartmentHistory] ed
  ON e.BusinessEntityID = ed.BusinessEntityID
  INNER JOIN [AdventureWorks2014].[HumanResources].[Department] d
  ON ed.DepartmentID = d.DepartmentID

--Merge Join
SELECT SOH.SalesOrderID, SOD.ProductID, SOD.OrderQTY
  FROM [AdventureWorks2014].[Sales].[SalesOrderDetail] SOD
  INNER JOIN [AdventureWorks2014].[Sales].[SalesOrderHeader] SOH
  ON SOD.SalesOrderID = SOH.SalesOrderID


--Computer Scalar
SELECT FirstName + ' ' + LastName
FROM [AdventureWorks2014].[Person].[Person]

--Filter
--Missing index
SELECT * 
FROM [AdventureWorks2014].[Sales].[SalesOrderDetail]
WHERE OrderQTY > 10

--Sort
SELECT * 
FROM [AdventureWorks2014].[Sales].[SalesOrderDetail]
ORDER BY OrderQty

--Top
SELECT TOP 10 * 
FROM [AdventureWorks2014].[Sales].[SalesOrderDetail]


--Comparing Two queries
--Forcing Hints
SELECT SOH.SalesOrderID, SOD.ProductID, SOD.OrderQTY
  FROM [AdventureWorks2014].[Sales].[SalesOrderDetail] SOD
  INNER JOIN [AdventureWorks2014].[Sales].[SalesOrderHeader] SOH
  ON SOD.SalesOrderID = SOH.SalesOrderID

SELECT SOH.SalesOrderID, SOD.ProductID, SOD.OrderQTY
  FROM [AdventureWorks2014].[Sales].[SalesOrderDetail] SOD
  INNER LOOP JOIN [AdventureWorks2014].[Sales].[SalesOrderHeader] SOH
  ON SOD.SalesOrderID = SOH.SalesOrderID

---- New Features
-- Compare Showplan
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName LIKE N'%Sanchez'

/* Compare with:
SELECT FirstName, MiddleName, LastName 
FROM [AdventureWorks2014].[Demo].[Person] 
WHERE LastName LIKE N'Sanchez%'
*/

-- Cardinality Estimator

SELECT SOH.SalesOrderID, SOD.ProductID, SOD.OrderQTY
  FROM [AdventureWorks2014].[Sales].[SalesOrderDetail] SOD
  INNER JOIN [AdventureWorks2014].[Sales].[SalesOrderHeader] SOH
  ON LTRIM(SOD.SalesOrderID) = LTRIM(SOH.SalesOrderID)
  INNER JOIN [AdventureWorks2014].[Production].Product P 
  ON LTRIM(SOD.ProductID) = LTRIM(P.ProductID)
  INNER JOIN [AdventureWorks2014].[Production].ProductSubcategory PS
  ON LTRIM(P.ProductSubcategoryID) = LTRIM(PS.ProductSubcategoryID)
  WHERE SOD.SalesOrderID = 43659

SELECT SOH.SalesOrderID, SOD.ProductID, SOD.OrderQTY
  FROM [AdventureWorks2014].[Sales].[SalesOrderDetail] SOD
  INNER JOIN [AdventureWorks2014].[Sales].[SalesOrderHeader] SOH
  ON LTRIM(SOD.SalesOrderID) = LTRIM(SOH.SalesOrderID)
  INNER JOIN [AdventureWorks2014].[Production].Product P 
  ON LTRIM(SOD.ProductID) = LTRIM(P.ProductID)
  INNER JOIN [AdventureWorks2014].[Production].ProductSubcategory PS
  ON LTRIM(P.ProductSubcategoryID) = LTRIM(PS.ProductSubcategoryID)
  WHERE SOD.SalesOrderID = 43659
  OPTION (USE HINT ('FORCE_LEGACY_CARDINALITY_ESTIMATION'))

-- Live Query Statistics
/*
SELECT SOH.SalesOrderID, SOD.ProductID, SOD.OrderQTY
  FROM [AdventureWorks2014].[Sales].[SalesOrderDetail] SOD
  CROSS JOIN [AdventureWorks2014].[Sales].[SalesOrderHeader] SOH
  */
 
 -- Row Level Security
GO
Create Function fn_firstletter (@empname sysname)
returns table
with Schemabinding
as
return select 1 as [result]
where @empname = 'Mason' --Normally would be filtered on username

GO
--Table Scan
SELECT BusinessEntityID FROM AdventureWorks2014.Demo.Person

GO

CREATE SECURITY POLICY NameFilter  
ADD FILTER PREDICATE dbo.fn_firstletter(FirstName)   
ON Demo.Person
WITH (STATE = ON);

--Index Scan
SELECT BusinessEntityID FROM AdventureWorks2014.Demo.Person

SELECT FirstName, BusinessEntityID FROM AdventureWorks2014.Demo.Person



--Clean up
DROP SECURITY POLICY IF EXISTS [dbo].[NameFilter]
DROP FUNCTION IF EXISTS [dbo].[fn_firstletter]
DROP TABLE IF EXISTS [Demo].[Person]